# -*- coding: utf-8 -*-
"""
Created on Thu Mar 03 05:17:18 2016

@author: Han-Wei
"""

from __future__ import print_function
import csv
import const
import numpy as np
from linkedlist import Node, LinkedList

(LATITUDE, LONGITUDE, YEAR, SEX) = range(4)


def evalSex(sexStr):
    if sexStr == 'Male':
        return 1.0
    elif sexStr == 'Female':
        return 0.0
    else:
        raise ValueError('Invalid string')


def load_csv(fileName):
    minLati = const.LARGEST_LATITUDE
    maxLati = const.SMALLEST_LATITUDE
    minLongi = const.LARGEST_LONGITUDE
    maxLongi = const.SMALLEST_LONGITUDE
    
    with open(fileName, 'rb') as inFile:
        hasHeader = csv.Sniffer().has_header(inFile.read(1024))
        inFile.seek(0)  # rewind
        inCSV = csv.reader(inFile)
        if hasHeader:
            next(inCSV)  # skip header row
        ls = LinkedList()
        
        rowCount = 2
        for row in inCSV:
            
            if row[LATITUDE]:
                lati = eval(row[LATITUDE])
            else:
                raise ValueError('Row ' + str(rowCount) + ' has a blank field in LATITUDE.')
                
            if row[LONGITUDE]:
                longi = eval(row[LONGITUDE])
            else:
                raise ValueError('Row ' + str(rowCount) + ' has a blank field in LONGITUDE.')
            
            if row[YEAR]:
                year = eval(row[YEAR])
            else:
                raise ValueError('Row ' + str(rowCount) + ' has a blank field in YEAR.')
             
            if row[SEX]:
                sex = evalSex(row[SEX])
            else:
                raise ValueError('Row ' + str(rowCount) + ' has a blank field in SEX.')
            
            ls.add_tail([lati, longi, year, sex])
            
            if lati < minLati:
                minLati = lati
            if lati > maxLati:
                maxLati = lati
            if longi < minLongi:
                minLongi = longi
            if longi > maxLongi:
                maxLongi = longi
                
            rowCount += 1
            
    return ls, minLati, maxLati, minLongi, maxLongi
        
        
def compute_bucket_size(minLati, maxLati, minLongi, maxLongi):
    height = abs(maxLati - minLati)
    width = abs(maxLongi - minLongi)
    
    if width > height:
        bucketSize = width / const.NUM_BUCKET
    else:
        bucketSize = height / const.NUM_BUCKET
        
    return bucketSize


def compute_boundary(bucketSize, minLati, maxLati, minLongi, maxLongi):
    height = abs(maxLati - minLati)
    width = abs(maxLongi - minLongi)
    
    if width > height:
        newMaxLati = maxLati + (width - height) / 2.0
        newMinLati = minLati - (width - height) / 2.0
        newMaxLongi = maxLongi
        newMinLongi = minLongi
    else:
        newMaxLongi = maxLongi + (height - width) / 2.0
        newMinLongi = minLongi - (height - width) / 2.0
        newMaxLati = maxLati
        newMinLati = minLati
        
    newMaxLati = newMaxLati + const.MARGIN_SIZE * bucketSize
    newMinLati = newMinLati - const.MARGIN_SIZE * bucketSize
    newMaxLongi = newMaxLongi + const.MARGIN_SIZE * bucketSize
    newMinLongi = newMinLongi - const.MARGIN_SIZE * bucketSize
    
    return newMinLati, newMaxLati, newMinLongi, newMaxLongi


def build_2d_array(ls, bucketSize, minLati, maxLati, minLongi, maxLongi):
    nrows = np.int_(np.rint((maxLati + bucketSize - minLati) / bucketSize))
    ncols = np.int_(np.rint((maxLongi + bucketSize - minLongi) / bucketSize))
    
    if not nrows == ncols:
        raise AssertionError("The 2D array is not square.") 
        
    arr2d = np.zeros((nrows, ncols), dtype=np.int32)
    
    entry = ls.head
    while entry != None:
        lati = entry.get_data()[LATITUDE]
        longi = entry.get_data()[LONGITUDE]
        
        i = np.int_(np.rint((longi - minLongi) / bucketSize))
        j = np.int_(np.rint((lati - minLati) / bucketSize))        
        
        arr2d[j, i] = arr2d[j, i] + 1
        entry = entry.get_next()
    return arr2d
    
    
def build_year_2d_array(ls, bucketSize, minLati, maxLati, minLongi, maxLongi):
    nrows = np.int_(np.rint((maxLati + bucketSize - minLati) / bucketSize))
    ncols = np.int_(np.rint((maxLongi + bucketSize - minLongi) / bucketSize))
    
    if not nrows == ncols:
        raise AssertionError("The 2D array is not square.")     
    matrix = [[LinkedList() for i in range(ncols)] for j in range(nrows)]
    
    entry = ls.head
    while entry != None:
        lati = entry.get_data()[LATITUDE]
        longi = entry.get_data()[LONGITUDE]
        
        i = np.int_(np.rint((longi - minLongi) / bucketSize))
        j = np.int_(np.rint((lati - minLati) / bucketSize))        
        
        matrix[j][i].add_tail(entry.get_data()[YEAR])
        entry = entry.get_next()
    
    arr2d = np.zeros((nrows, ncols), dtype=np.float64)    
    for j in range(nrows):
        for i in range(ncols):            
            size = matrix[j][i].size()
            if size != 0:
                entry = matrix[j][i].head
                total = 0.0                
                while entry != None:
                    total = total + entry.get_data()
                    entry = entry.get_next()                
                arr2d[j, i] = total / size
                # print(str(arr2d[j, i]) + ', ', end="")
            #else:
                # print('0, ', end="")
        # print("\n")
    return arr2d

def normalize(arr2d):
    ones2d = np.ones(arr2d.shape)
    newArr2d = ones2d + arr2d
    newArr2d = np.log10(newArr2d)
    maximum = np.amax(newArr2d)
    return newArr2d / maximum


def normalizeYear(yearArr2d):
    # ones2d = np.ones(yearArr2d.shape)
    # newArr2d = ones2d + yearArr2d
    # newArr2d = np.log10(newArr2d)
    minYear = const.LARGEST_YEAR
    maxYear = const.SMALLEST_YEAR
    # return newArr2d / maximum

    nrows, ncols = arr2d.shape
    
    for j in range(nrows):
        for i in range(ncols):
            if yearArr2d[j, i] != 0:
                if yearArr2d[j, i] < minYear:
                    minYear = yearArr2d[j, i]
                if yearArr2d[j, i] > maxYear:
                    maxYear = yearArr2d[j, i]
    # print(minYear)
    # print(maxYear)
    
    for j in range(nrows):
        for i in range(ncols):
            if yearArr2d[j, i] != 0:
                yearArr2d[j, i] = (yearArr2d[j, i] - minYear) / (maxYear - minYear)
                
    return yearArr2d
    
    
def write_csv(fileName, arr2d, yearArr2d):
    nrows, ncols = arr2d.shape
    
    xs = np.linspace(0, 1, ncols)
    ys = np.linspace(0, 1, nrows)    
    
    with open(fileName, 'wb') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(['x'] + ['y'] + ['z'] + ['year']) 
        for j in range(nrows):
            for i in range(ncols):
                writer.writerow([xs[i]] + [ys[j]] + [arr2d[j][i]] + [yearArr2d[j][i]])


if __name__ == '__main__':
    
    bugList, minLati, maxLati, minLongi, maxLongi = \
    load_csv('occurrences.csv')
    
    bucketSize = \
    compute_bucket_size(minLati, maxLati, minLongi, maxLongi)
    
    minLati, maxLati, minLongi, maxLongi = \
    compute_boundary(bucketSize, minLati, maxLati, minLongi, maxLongi)         
    
    arr2d = build_2d_array(bugList, bucketSize, minLati, maxLati, minLongi, maxLongi)
    
    # print(arr2d)
    
    arr2d = normalize(arr2d)
    
    yearArr2d = build_year_2d_array(bugList, bucketSize, minLati, maxLati, minLongi, maxLongi)
    
    yearArr2d = normalizeYear(yearArr2d)
    
    # print(yearArr2d)
    
    write_csv("data.csv", arr2d, yearArr2d)
    
    """
    print (bucketSize, minLati, maxLati, minLongi, maxLongi)
    """