import peasy.*;
import peasy.org.apache.commons.math.geometry.*;
// import controlP5.*;

PeasyCam cam;
CameraState camFrontState;

boolean rotateStatus = false;
boolean wireframeStatus = true;
boolean shadingStatus = true;

int iteration = 4;
int roughnessFactor = 8;

final int maxRoughnessFactor = 16;
final int minRoughnessFactor = 1;

final int n = 13;
final int baseSize = 13;

final int X = 0;
final int Y = 1;
final int Z = 2;
final int YEAR = 3;

final float scale = 1500;
final float rotFactor = 0.05;  // Radians rotation per frame

My3DArray list;
float[][] baseArray;
float[][] colorArray;
My3DArray map;
My3DArray colorMap;
PShape[] strips;

// ControlP5 cp5;

void setup() {
  size(1600, 960, P3D);
  cam = new PeasyCam(this, 2500);
  cam.setMinimumDistance(.1);
  cam.setMaximumDistance(5000);
  
  camFrontState = setCamFrontView(cam);
  
  // cp5 = new ControlP5(this);
  
  list = load_csv("data.csv");
  
  baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
  colorArray = buildColorArray(baseSize, list, roughnessFactor);
  
  map = array2DToMy3Darray(baseArray);  
  colorMap = array2DToMy3Darray(colorArray);  
  
  buildStrips(map, colorMap);  
  
  // add a horizontal slider
  /*
  cp5.addSlider("iteration_slider")
     .setPosition(width / 4, height / 7.5)
     .setSize(200,20)
     .setRange(0, maxIterations)
     .setValue(3)
     .setNumberOfTickMarks(maxIterations + 1);
     ;
  cp5.setAutoDraw(false);
  */ //<>// //<>//
}


void draw() {
  background(0);
  
  // buildStrips(map, colorMap);
  pushMatrix();
  if(rotateStatus){
    cam.rotateY(rotFactor);
  }
  
  translate(-0.5 * scale, -0.5 * scale, 0);    
  
  for (PShape s: strips) {
    shape(s);
  }
  popMatrix();
  drawGui();
}


void keyReleased() {
  if (key != CODED) {
    if (key == 'f' || key == 'F') {
      cam.setState(camFrontState);
    }
    if (key == 'r' || key == 'R') {
      rotateStatus = !rotateStatus;
    }
    if (key == 'w' || key == 'W') {
      wireframeStatus = !wireframeStatus;
      buildStrips(map, colorMap);
      // iteration_slider(iteration_slider);
    }
    if (key == 's' || key == 'S') {
      shadingStatus = !shadingStatus;
      buildStrips(map, colorMap);
      // iteration_slider(iteration_slider);
    }  
    
    if (key == '0' || key == ')') {
      iteration = 0;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }  
    
    if (key == '1' || key == '!') {
      iteration = 1;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }
    
    if (key == '2' || key == '@') {
      iteration = 2;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }
    
    if (key == '3' || key == '#') {
      iteration = 3;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }
    
    if (key == '4' || key == '$') {
      iteration = 4;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }
    
    if (key == '5' || key == '%') {
      iteration = 5;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }  
    
    if (key == '6' || key == '^') {
      iteration = 6;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }
    
    if (key == '>' || key == '.') {
      println("Right arrow key pressed.");
      if(roughnessFactor > minRoughnessFactor) {
        
        roughnessFactor -= 1;
        println("Roughness: " + roughnessFactor);
        baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
        colorArray = buildColorArray(baseSize, list, roughnessFactor);
        
        map = array2DToMy3Darray(baseArray);  
        colorMap = array2DToMy3Darray(colorArray);
        
        buildStrips(map, colorMap);
      }
    }
    
    if (key == '<' || key == ',') {
      println("Left arrow key pressed.");
      if(roughnessFactor < maxRoughnessFactor) {
        roughnessFactor += 1;
        
        baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
        colorArray = buildColorArray(baseSize, list, roughnessFactor);
        
        map = array2DToMy3Darray(baseArray);  
        colorMap = array2DToMy3Darray(colorArray);
        
        buildStrips(map, colorMap);
      }
      else {
      }
      println("Roughness: " + roughnessFactor);
    }    
  }
  else {
    // println("Key is not coded.");
    if (key == RIGHT) {
      println("Left arrow key pressed.");
      if(roughnessFactor <= maxRoughnessFactor)
        roughnessFactor += 1;
      baseArray = buildBaseArray(baseSize, list, roughnessFactor);  
      colorArray = buildColorArray(baseSize, list, roughnessFactor);
      
      map = array2DToMy3Darray(baseArray);  
      colorMap = array2DToMy3Darray(colorArray);
      
      buildStrips(map, colorMap);
    }    
  }
}


void buildStrips(My3DArray map, My3DArray colorMap)
{
  colorMode(HSB, 1, 1, 1);
  strips = new PShape[map.get2ndDimensionSize() - 1];
  
  for(int j = 0; j < map.get2ndDimensionSize() - 1; j ++) {
    strips[j] = createShape();    
    strips[j].beginShape(TRIANGLE_STRIP);
    if(!wireframeStatus) {
      strips[j].noStroke();
    }
    for(int i = 0; i < map.get1stDimensionSize(); i ++) {
      float x = scale * (float)map.getElementIn3DArray(i, j, X);
      float y = scale * (float)map.getElementIn3DArray(i, j, Y);
      float z = 0.125 * scale * (float)map.getElementIn3DArray(i, j, Z);
      float c = (float)colorMap.getElementIn3DArray(i, j, Z);
      strips[j].vertex(x, y, z);
      if(shadingStatus) {
        c = map(c, 0, 1, 0.083333, 0.5);
        strips[j].fill(c, 1, 1 - c);
      }
      
      
      x = scale * (float)map.getElementIn3DArray(i, j + 1, X);
      y = scale * (float)map.getElementIn3DArray(i, j + 1, Y);
      z = 0.125 * scale * (float)map.getElementIn3DArray(i, j + 1, Z);
      c = (float)colorMap.getElementIn3DArray(i, j + 1, Z);
      strips[j].vertex(x, y, z);
      if(shadingStatus) {
        c = map(c, 0, 1, 0.083333, 0.5);
        strips[j].fill(c, 1, 1 - c);
      }
    }
    strips[j].endShape(CLOSE);
  }  
}


void drawGui() {
  hint(DISABLE_DEPTH_TEST);
  cam.beginHUD();
  drawLegend();
  // cp5.draw();
  cam.endHUD();
  hint(ENABLE_DEPTH_TEST);
}


void drawLegend()
{
  int fontSize = 24;
  textSize(fontSize);
  textAlign(LEFT, TOP);
  text("1800", 0, height / 20);
  text("Year collected", 0, fontSize + height / 20);
  textAlign(RIGHT, TOP);
  text("2010", width, height / 20);
  
  textAlign(CENTER, BOTTOM);
  text("The Occurrence of Plant Bugs in North America", width / 2, height);  
  
  // noStroke();
  colorMode(HSB, 1);
  int barWidth = width;
  int barHeight = height / 20;
  for (int i = 0; i < barWidth; i++) {
    for (int j = 0; j < barHeight; j++) {
      float hue = map(i, 0, barWidth, 0.083333, 0.5);
      float val = 1 - hue;
      stroke(hue, 1, val);
      point(i, j);
    }
  }
  stroke(0);          // Setting the outline (stroke) to black 
}

/*
void iteration_slider(int theIteration) {  
  println("a slider event. setting iterations to " + theIteration);
  if(mousePressed)
  {
    iteration_slider = theIteration;
    baseArray = buildBaseArray(baseSize, list);  
    colorArray = buildColorArray(baseSize, list);
    
    map = array2DToMy3Darray(baseArray);  
    colorMap = array2DToMy3Darray(colorArray);
    
    buildStrips(map, colorMap);
  }
}
*/